import numpy as np
import scipy.stats as ss
import matplotlib.pyplot as plt
import stan
from my_module import *
import pandas as pd
import arviz as az

pd.set_option('display.max_columns', 10000)
pd.set_option('display.max_rows', 10000)
   
fin_nm = input('Data file name (*.xlsx) = ')
fout_nm = input('Output file name (*.txt) = ')
fout = open(fout_nm, 'w')
fout.write('Data file...{}\n'.format(fin_nm))
#
#       Read data from Excel file
#
xlsx_fl = pd.ExcelFile(fin_nm)
data_xlsx = pd.read_excel(xlsx_fl,  header = None)
data = data_xlsx.values
var_name = data[0]
values = data[1:]
N = len(values)
ID = []
X = np.empty(N, dtype = float)
Res = np.empty(N, dtype = int)
for t in range(N):
    ID.append(values[t][0])
    X[t] = float(values[t][1])
    Res[t] = int(values[t][2])

print('Number of data = ', N)
fout.write('\nNumber of data = {}\n'.format(N))
for i in range(N):
    print('{0}  {1}  {2}'.format(ID[i], X[i], Res[i]))

for i, s, r in zip(ID, X, Res):
    fout.write('{0:>5}{1:>7.2f}{2:>5d}\n'.format(i, s, r))
   
#
#       Values for rescaling in MCMC sampling
# 
mid_x, max_x = np.percentile(X, [50, 95])

def f_init():
    """
                Initial values for v_C, v_mu, and v_sigma
    """
    return dict(v_C = 0.5, v_mu = 0.0, v_sgm = 1.0)

stan_data = {'N': N, 'X':X, 'Res': Res, 'mid_x': mid_x, 'max_x': max_x}
#
#       Sampling by Stan
#
with open('PM4CatPS3.stan', 'r') as f_stan:
    sm = stan.build(f_stan.read(), data = stan_data)
    fit = sm.sample(num_chains = 4, init = [f_init()]*4)

dframe = fit.to_frame()
i_data = az.from_pystan(posterior = fit, posterior_model = sm)

print(az.summary(i_data, var_names = ['mu', 'sgm', 'C']))
fout.write('\nSummary:\n')
fout.write("{0}\n".format(az.summary(i_data, var_names = ['mu', 'sgm', 'C'])))

az.plot_trace(i_data, var_names = ['mu', 'sgm', 'C'])
plt.tight_layout()
plt.show()

ck = az.plot_posterior(i_data, var_names = 'mu', point_estimate = 'median')
plt.show()

ck = az.plot_posterior(i_data, var_names = 'sgm', point_estimate = 'median')
plt.show()

ck = az.plot_posterior(i_data, var_names = 'C', point_estimate = 'median')
plt.show()

#
#      MAP estimates of parameters
#
muMAP = CalcMAPEst(dframe['mu'])[0]
sgmMAP = CalcMAPEst(dframe['sgm'])[0]
CMAP = CalcMAPEst(dframe['C'])[0]
#
#        2.5 and 97.5 percentiles of posterior distributions
#
MuL025, MuU975 = np.percentile(dframe['mu'], [2.5, 97.5])
SgmL025, SgmU975 = np.percentile(dframe['sgm'], [2.5, 97.5])

#
#       Prob(z < s75) = 0.75 for the standard normal distribution
#
s75 = ss.norm.ppf(0.75, 0, 1)
print('s75 = ', s75, '     p = ', f_phi(s75))   #   f_phi from my_module.py
JND = sgmMAP * s75
PSE = muMAP
fout.write('\nMAP for PSE = {0:<.2f},    95% CI for PSE = [{1:<.2f}, {2:<.2f}]'.format(
            PSE, MuL025, MuU975))
fout.write('\nMAP for JND = {0:<.2f},    95% CI for JND= [{1:<.2f}, {2:<.2f}]'.format(
            JND, s75 * SgmL025, s75 * SgmU975))
#
#       Data and the psychometric function
#
pf = PF(muMAP, sgmMAP)          #   Class PF from my_module.py
minX = X[np.argmin(X)]
maxX = X[np.argmax(X)]
xvalues = []
yvalues = []
yvalues_L = []
yvalues_R = []
for i in range(1001):
    t = minX + i * (maxX - minX) / 1000.0
    xvalues.append(t)
    yvalues.append(pf.v(t))
    yvalues_L.append(pf.v(t + CMAP))
    yvalues_R.append(pf.v(t - CMAP))
plt.title('Psychometric Functions\n' +
          ('PSE = {0:<.2f},  JND = {1:<.2f},  C = {2:<.2f}'.format(PSE, JND, CMAP)),
          fontsize = 16)
plt.plot(xvalues, yvalues_L, 'g-', linewidth = 3, label = 'PF(>1)')
plt.plot(xvalues, yvalues, 'k-', linewidth = 3, label = 'PF(>2)')
plt.plot(xvalues, yvalues_R, 'b-', linewidth = 3, label = 'PF(>3)')
plt.yticks([0.0, 0.5, 0.75, 1.0])
#
#       Responses are scaled to values from 0 to 1, and added small fluctuations
#
Res_rn = (Res - 1) / 3 + np.random.uniform(-0.03, 0.03, len(Res))
plt.scatter(X, Res_rn, alpha = 0.5, color = 'g', label = 'Data')
plt.plot([minX, maxX], [0.0, 0.0], color = 'k')
plt.plot([minX, maxX], [1.0, 1.0], color = 'k')
plt.plot([minX, PSE, PSE], [0.5, 0.5, 0.0], color = 'k')
plt.plot([minX, PSE + JND, PSE + JND], [0.75, 0.75, 0.0], color = 'k')
plt.legend(loc = 7)
plt.xlabel('Stimulus value', fontsize = 12)
plt.ylabel('Probability', fontsize = 12)
plt.show()

#
#       Observed responses smoothed by Kernel density estimation 
#
w_1 = np.empty(len(Res))        #   Weights for Response >= 2
w_2 = np.empty(len(Res))        #   Weights for response >= 3
w_3 = np.empty(len(Res))        #   Weights for response >= 4
for t in range(len(Res)):
    w_1[t] = 0 if Res[t] < 2 else 1
    w_2[t] = 0 if Res[t] < 3 else 1
    w_3[t] = 0 if Res[t] < 4 else 1

X_range = np.arange(minX, maxX, (maxX - minX) / 100)
w_k = JND / 2

while True:
    pm_1 = []           #   KDE values for Response >= 2
    pm_2 = []           #   KDE values for Response >= 3
    pm_3 = []           #   KDE values for Response >= 4
    for v in X_range:
        pm_1.append(k_ave(v, X, w_1, w_k))
        pm_2.append(k_ave(v, X, w_2, w_k))
        pm_3.append(k_ave(v, X, w_3, w_k))
        
    plt.title('Observed and Estimated PFs\n' +
              ('PSE = {0:<.2f},  JND = {1:<.2f},  C = {2:<.2f}'.format(PSE, JND, CMAP)),
              fontsize = 16)
    plt.plot(xvalues, yvalues_L, 'g-', linewidth = 3, label = 'PF(>1)')
    plt.plot(xvalues, yvalues, 'k-', linewidth = 3, label = 'PF(>2)')
    plt.plot(xvalues, yvalues_R, 'b-', linewidth = 3, label = 'PF(>3)')
    plt.yticks([0.0, 0.5, 1.0])
    plt.plot(X_range, pm_1, 'g--', linewidth = 2, label = 'Res(>1)')
    plt.plot(X_range, pm_2, 'k--', linewidth = 2, label = 'Res(>2)')
    plt.plot(X_range, pm_3, 'b--', linewidth = 2, label = 'Res(>3)')
    plt.xlabel('Stimulus value', fontsize = 12)
    plt.legend(loc = 'upper left')
    plt.show()
    print('\nThe current value of w_k = {}'.format(w_k))
    w_k = float(input('The next value of w_k = '))
    if w_k <= 0.0:
        break

fout.close()
print('\nOutput file {0} was saved.\n'.format(fout_nm))

